#!/usr/bin/env python3
import unittest

from km3services._tools import _infinite, _zipequalize


def test_infinite():
    a = 1
    infinite_a = _infinite(a)
    for n in [1, 5, 100]:
        assert n == len(list(zip(range(n), _infinite(a))))


class TestZipEqualise(unittest.TestCase):
    def test_zipequalize(self):
        assert 1 == len(list(zip(*_zipequalize([1], [1]))))
        assert 3 == len(list(zip(*_zipequalize([1], [1, 2, 3]))))
        assert 3 == len(list(zip(*_zipequalize([1], [1, 2, 3], [4]))))

        a, b, c = list(zip(*_zipequalize([1], [1, 2, 3])))
        self.assertTupleEqual(a, (1, 1))
        self.assertTupleEqual(b, (1, 2))
        self.assertTupleEqual(c, (1, 3))

    def test_zipequalize_raises_if_dim_mismatch(self):
        with self.assertRaises(ValueError) as e:
            assert 3 == len(list(zip(*_zipequalize([1, 2], [1, 2, 3]))))



import json
import requests
import numpy as np


class NumpyEncoder(json.JSONEncoder):
    """ Custom encoder for numpy data types """

    def default(self, obj):
        if isinstance(
            obj,
            (
                np.int_,
                np.intc,
                np.intp,
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ),
        ):

            return int(obj)

        elif isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
            return float(obj)

        elif isinstance(obj, (np.complex_, np.complex64, np.complex128)):
            return {"real": obj.real, "imag": obj.imag}

        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()

        elif isinstance(obj, (np.bool_)):
            return bool(obj)

        elif isinstance(obj, (np.void)):
            return None

        return json.JSONEncoder.default(self, obj)


def _infinite(scalar):
    """Creates an infinite generator from a scalar.

    Useful when zipping finite length iterators with single scalars.
    """
    while True:
        yield scalar


def _zipequalize(*iterables):
    """Creates equal length iterables for a mix of single and n-length arrays"""

    dims = set(map(len, iterables))
    if len(dims) == 1:
        return iterables

    if 1 not in dims or len(dims) > 2:
        raise ValueError("Input arrays dimensions mismatch.")

    out = []
    for it in iterables:
        if len(it) == 1:
            out.append(_infinite(it[0]))
        else:
            out.append(it)

    return out



#!/usr/bin/env python3

import unittest
import numpy as np

import km3services


class TestOscProb(unittest.TestCase):
    def test_oscillationprobabilities(self):
        n = 100
        energies = np.random.randint(1, 50, n)
        cos_zeniths = 1 - np.random.rand(n) * 2
        probabilities = km3services.oscprob.oscillationprobabilities(
            12, 14, energies, cos_zeniths
        )
        assert np.all(probabilities <= 1)



import json
import requests
import numpy as np
import random

from typing import List, Optional

from fastapi import FastAPI, HTTPException
from pydantic import BaseModel


from km3services._tools import NumpyEncoder, _zipequalize


def oscillationprobabilities(
    flavour_in, flavour_out, energies, cos_zeniths, params=None
):
    """
    Calculate the neutrino oscillation properties.

    Parameters
    ----------
    flavour_in: float or array(float)
      PDG ID of the initial neutrino.
    flavour_out: float or array(float)
      PDG ID of the output neutrino.
    energies: float or array(float)
      The neutrino energies in GeV
    cos_zenith: float [-1; 0] or array(float)
      The cosine of the zenith (-1 is upgoing)
    params: dict(str: float)
      Oscillation parameters (dm_21, dm_31, theta_12, theta_13, dcp)

    Returns
    -------
    array(float)

    """
    if params is None:
        params = {
            "dm_21": 7.40e-5,
            "dm_31": 2.494e-3,
            "theta_12": 5.868e-1,
            "theta_23": 8.238e-1,
            "theta_13": 1.491e-1,
            "dcp": 4.084070449666731,
        }

    data = {
        "params": params,
        "nus": {
            "flavour_in": np.atleast_1d(flavour_in),
            "flavour_out": np.atleast_1d(flavour_out),
            "energies": np.atleast_1d(energies),
            "cos_zeniths": np.atleast_1d(cos_zeniths),
        },
    }
    r = requests.post(
        "http://131.188.167.67:30000/probabilities",
        data=json.dumps(data, cls=NumpyEncoder),
    )
    if r.ok:
        return np.array(json.loads(r.text))
    else:
        if r.status_code == 460:  # TODO: standardise this magic number
            raise ValueError(json.loads(r.text)["detail"])
        print(f"Error: {r.reason}")
        print(r.text)


# Server part

app = FastAPI()


class OscillationParameters(BaseModel):
    dm_21: float
    dm_31: float
    theta_12: float
    theta_23: float
    theta_13: float
    dcp: float


class _NumpyArray(BaseModel):
    dtype: str
    data: str


def tonumpy(arr: _NumpyArray):
    return np.frombuffer(arr.data)


class _Neutrinos(BaseModel):
    flavour_in: List[int]
    flavour_out: List[int]
    energies: List[float]
    cos_zeniths: List[float]


# @app.post("/numpy")
# async def _numpy(array: _NumpyArray):
#     return tonumpy(array)


@app.post("/probabilities")
async def _probability(params: OscillationParameters, nus: _Neutrinos):
    import ROOT

    ROOT.gSystem.Load("libOscProb.so")
    pmns = ROOT.OscProb.PMNS_Fast()
    prem = ROOT.OscProb.PremModel()

    pmns.SetDm(2, params.dm_21)  # set delta_m21 in eV^2
    pmns.SetDm(3, params.dm_31)  # set delta_m31 in eV^2
    pmns.SetAngle(1, 2, params.theta_12)  # set Theta12 in radians
    pmns.SetAngle(1, 3, params.theta_13)  # set Theta13 in radians
    pmns.SetAngle(2, 3, params.theta_23)  # set Theta23 in radians
    pmns.SetDelta(1, 3, params.dcp)  # set Delta13 in radians

    try:
        p = _osc_prob(
            pmns,
            prem,
            nus.flavour_in,
            nus.flavour_out,
            nus.energies,
            nus.cos_zeniths,
        )
    except ValueError as e:
        raise HTTPException(status_code=460, detail=str(e))

    return p


def _osc_prob(pmns, prem, flav_in, flav_out, energies, cos_zenith):
    params = _zipequalize(flav_in, flav_out, energies, cos_zenith)

    # use a PremModel to make the paths through the earth
    # with the class PremModel
    # chose an angle for the neutrino and fill the paths with cosTheta,
    # e.g. cosTheta = -1 (vertical up-going)
    P = []
    for fl_in, fl_out, E, cosZ in zip(*params):
        if fl_in < 0:
            pmns.SetIsNuBar(True)
        else:
            pmns.SetIsNuBar(False)
        prem.FillPath(cosZ)
        pmns.SetPath(prem.GetNuPath())
        P.append(pmns.Prob(_pdgid2flavor(fl_in), _pdgid2flavor(fl_out), E))
    return P


def _pdgid2flavor(pdgid):
    """Converts PDG ID to OscProb flavor"""
    if abs(pdgid) == 12:
        return 0
    if abs(pdgid) == 14:
        return 1
    if abs(pdgid) == 16:
        return 2
    raise ValueError("Unsupported PDG ID, please use neutrinos")
